@@ -6,6 +6,7 @@ use feature qw(state);
 use utf8;
 
 use Moo;
+use Data::Util qw(:check);
 use Data::Compare;
 use RDF::NS;
 use RDF::Trine;
@@ -16,17 +17,24 @@ use HTTP::Request::Common;
 use Log::Any ();
 use Cache::LRU;
 use Clone qw(clone);
+use Data::Dumper;
 use JSON;
-use URI::Template;
-use RDF::LDF::Error;
 
-our $VERSION = '0.23_01';
+our $VERSION = '0.01';
 
 has url => (
     is => 'ro' ,
     required => 1
 );
 
+has ua => (
+    is      => 'ro',
+    lazy    => 1,
+    builder => sub {
+        LWP::UserAgent->new( agent => "RDF:::LDF/$RDF::LDF::VERSION" );
+    }
+);
+
 has sn => (
     is     => 'ro' ,
     lazy   => 1,
@@ -35,6 +43,12 @@ has sn => (
     }
 );
 
+has pattern => (
+    is      => 'ro',
+    lazy    => 1,
+    builder => 'get_pattern'
+);
+
 has lru => (
     is     => 'ro' ,
     lazy   => 1,
@@ -44,68 +58,39 @@ has lru => (
 );
 
 has log => (
-    is    => 'ro',
-    lazy  => 1,
-    builder => sub {
-        Log::Any->get_logger(category => ref(shift));
-    }
+	is    => 'ro',
+	lazy  => 1,
+	builder => sub {
+		Log::Any->get_logger(category => ref(shift));
+	}
 );
 
-# Public method
 sub is_fragment_server {
-    my $self = shift;
-    my $federated  = ref($self->url) ? $self->url : [ $self->url ];
-    for my $part (@$federated) {
-        return 0 unless $self->get_query_pattern($part);
-    }
-    return 1;
+    shift->pattern ? 1 : 0;
 }
 
-# Public method
-# Optimized method to find all bindings matching a pattern
-# See:
-# Verborgh, Ruben, et al. Querying Datasets on the Web with High Availability. ISWC2014
-# http://linkeddatafragments.org/publications/iswc2014.pdf
-sub get_pattern {
-    my ($self,$bgp,$context,%args) = @_;
+sub get_sparql {
+    my ($self,$sparql) = @_;
 
-    unless (defined $bgp) {
-       RDF::LDF::Error->throw(text => "can't execute get_pattern for an empty pattern");
+    $self->log->debug("parsing $sparql");
+    my $query = RDF::Query->new($sparql);
+
+    unless (defined $query) {
+        $self->log->error("failed to parse $sparql");
+        return undef;
     }
 
-    my (@triples)   = ($bgp->isa('RDF::Trine::Statement') or $bgp->isa('RDF::Query::Algebra::Filter'))
-                    ? $bgp
-                    : $bgp->triples;
+    my $bgps = $self->_parse_bgp($query->pattern);
 
-    unless (@triples) {
-        RDF::LDF::Error->throw(text => "can't execute get_pattern for an empty pattern");
+    unless (is_array_ref($bgps)) {
+        $self->log->error("can't parse bgp for sparql");
+        return undef;
     }
 
-    my @vars = $bgp->referenced_variables;
-
-    my @bgps = map { $self->_parse_triple_pattern($_)} @triples;
-
-    my $sub = sub {
-        state $it = $self->_find_variable_bindings(\@bgps);
-        my $b = $it->();
-
-        return undef unless $b;
-
-        my $binding = RDF::Trine::VariableBindings->new({});
-
-        for my $key (keys %$b) {
-            my $val = $b->{$key};
-            $key =~ s{^\?}{};
-            $binding->set($key => $val);
-        }
-
-        $binding;
-    };
-
-    RDF::Trine::Iterator::Bindings->new($sub,\@vars);
+    $self->find_variable_bindings($bgps);
 }
 
-sub _find_variable_bindings {
+sub find_variable_bindings {
     my $self     = shift;
     my $bgps     = shift;
     my $bindings = shift // {};
@@ -116,11 +101,11 @@ sub _find_variable_bindings {
 
         my $ret;
 
-        # Loop over all variabe bindings with multiple matches
+        # Loop over all variabe bindinfgs with multiple matches
         while (!defined($ret = $results->())) {
             unless (defined $it) {
                 # Find the an binding iterator for the best pattern from $bgpgs
-                ($it,$bgps) = $self->_find_variable_bindings_($bgps);
+                ($it,$bgps) = $self->_find_variable_bindings($bgps);
 
                 return undef unless $it;
             }
@@ -137,9 +122,9 @@ sub _find_variable_bindings {
             # Apply all the bindings to the rest of the bgps;
             my $bgps_prime = $self->_apply_binding($this_binding,$bgps);
 
-            $results = $self->_find_variable_bindings($bgps_prime,$bindings);
+            $results = $self->find_variable_bindings($bgps_prime,$bindings);
         }
-
+        
         $ret;
     };
 
@@ -152,15 +137,15 @@ sub _find_variable_bindings {
 #  my ($iterator, $rest) = $self->_find_variable_bindings([ {pattern} , {pattern} , ... ]);
 #
 #  where:
-#
+# 
 #  $iterator - Iterator for variable bindings for the winnnig pattern, or undef when no
 #              patterns are provided or we get zero results
-#
+#              
 #  $rest     - An array ref of patterns not containing the best pattern
-sub _find_variable_bindings_ {
+sub _find_variable_bindings {
     my ($self,$bgps) = @_;
 
-    return (undef, undef) unless _is_array_ref($bgps) && @$bgps > 0;
+    return (undef, undef) unless is_array_ref($bgps) && @$bgps > 0;
 
     my ($pattern,$rest) = $self->_find_best_pattern($bgps);
 
@@ -186,7 +171,9 @@ sub _find_variable_bindings_ {
 
         for (keys %var_map) {
             my $method   = $var_map{$_};
-            $var_map{$_} = $triple->$method;
+            my $value    = $triple->$method->as_string;
+            $value =~ s{^<(.*)>$}{$1};
+            $var_map{$_} = $value;
         }
 
         return {%var_map};
@@ -198,7 +185,7 @@ sub _find_variable_bindings_ {
 sub _apply_binding {
     my ($self,$binding,$bgps) = @_;
 
-    return unless _is_array_ref($bgps) && @$bgps > 0;
+    return unless is_array_ref($bgps) && @$bgps > 0;
 
     my $copy = clone $bgps;
     my @new  = ();
@@ -206,10 +193,7 @@ sub _apply_binding {
     for my $pattern (@$copy) {
         for (qw(subject predicate object)) {
             my $val = $pattern->{$_};
-            if (defined($val) && $binding->{$val}) {
-                my $str_val    = $self->_node_as_string($binding->{$val});
-                $pattern->{$_} = $str_val
-            }
+            $pattern->{$_} = $binding->{$val} if defined($val) && $binding->{$val};
         }
         push @new, $pattern;
     }
@@ -217,7 +201,7 @@ sub _apply_binding {
     return \@new;
 }
 
-# Create a pattern which binds to the graph pattern
+# Create a pattern which binds optiomally to the graph pattern
 #
 # Usage:
 #
@@ -230,9 +214,9 @@ sub _apply_binding {
 #
 #    my ($pattern, $rest) = $self->_find_best_pattern($triples);
 #
-#    $pattern => Pattern in $triples which least amount of results
-#    $rest    => All patterns in $triples except $pattern
-#
+#    $pattern => Pattern in $triples which least ammount of results
+#    $rest    => All patterns in $triples except $pattern      
+#  
 sub _find_best_pattern {
     my ($self,$triples) = @_;
 
@@ -258,7 +242,7 @@ sub _find_best_pattern {
             $best_count   = $count;
             $best_pattern = $pattern;
         }
-    }
+    }   
 
     return (undef,$triples) unless defined $best_pattern;
 
@@ -267,14 +251,14 @@ sub _find_best_pattern {
     return ($best_pattern, \@rest_triples);
 }
 
-# Find the total number of triples available for a pattern
+# Find the total number of trilpes available for a pattern
 #
 # Usage:
 #
 #    my $count = $self->_total_triples(
 #                { subject => ... , predicate => ... , object => ...}
 #                );
-# Where
+# Where 
 #       $count is a number
 sub _total_triples {
     my ($self,$pattern) = @_;
@@ -289,46 +273,58 @@ sub _total_triples {
     $info->{hydra_totalItems};
 }
 
-sub _node_as_string {
-    my $self    = shift;
-    my $node    = shift;
-    if (_is_blessed($node) && $node->isa('RDF::Trine::Node')) {
-        if ($node->isa('RDF::Trine::Node::Variable')) {
-            return $node->as_string; # ?foo
-        } elsif ($node->isa('RDF::Trine::Node::Literal')) {
-            return $node->as_string; # includes quotes and any language or datatype
-        } else {
-            return $node->value; # the raw IRI or blank node identifier value, without other syntax
+# For an parsed SPARQL query find all BGP triples
+sub _parse_bgp {
+    my ($self,@pattern) = @_;
+
+    for my $p (@pattern) {
+        if ($p->isa('RDF::Query::Algebra::GroupGraphPattern')) {
+             my @triples = ();
+             for my $bgp ($p->patterns) {
+                push @triples , map { $self->_parse_triple_pattern($_)} $bgp->triples;
+             }
+             return \@triples;
+        }
+        else {
+            return $self->_parse_bgp($p->pattern);
         }
     }
-    return '';
+
+    return undef;
 }
 
 # For an BGP triple create a fragment pattern
 sub _parse_triple_pattern {
     my ($self,$triple) = @_;
-    my $subject   = $self->_node_as_string($triple->subject);
-    my $predicate = $self->_node_as_string($triple->predicate);
-    my $object    = $self->_node_as_string($triple->object);
-    my $hash      = {
+    my ($subject,$predicate,$object);
+
+    $subject   = $triple->subject->as_string;
+    $subject   =~ s{<(.*)>}{$1};
+
+    $predicate = $triple->predicate->as_string;
+    $predicate =~ s{<(.*)>}{$1};
+
+    $object    = $triple->object->as_string;
+    $object    =~ s{<(.*)>}{$1};
+
+    return {
         subject   => $subject ,
-        predicate => $predicate,
+        predicate => $predicate , 
         object    => $object
     };
-    return $hash;
 }
 
-# Dynamic find out which triple patterns need to be used to query the fragment server
+# Dynamic find out which tripple patterns need to be used to query the fragment server
 # Returns a hash:
 # {
 #   rdf_subject   => <name_of_subject_variable> ,
 #   rdf_predicate => <name_of_predicate_variable> ,
 #   rdf_object    => <name_of_object_variable>
-#   hydra_template => <endpoint_for_tripple_pattern>
+#   void_uriLookupEndpoint => <endpoint_for_tripple_pattern>
 # }
-sub get_query_pattern {
-    my ($self,$url) = @_;
-
+sub get_pattern {
+    my ($self) = @_;
+    my $url      = $self->url;
     my $fragment = $self->get_model_and_info($url);
 
     return undef unless defined $fragment;
@@ -337,12 +333,12 @@ sub get_query_pattern {
 
     my $pattern;
 
-    return undef unless _is_hash_ref($info);
+    return undef unless is_hash_ref($info);
 
-    return undef unless $info->{hydra_template};
+    return undef unless $info->{void_uriLookupEndpoint};
 
     for (keys %$info) {
-        next unless _is_hash_ref($info->{$_}) && $info->{$_}->{hydra_property};
+        next unless is_hash_ref($info->{$_}) && $info->{$_}->{hydra_property};
         my $property = join "_" , $self->sn->qname($info->{$_}->{hydra_property});
         my $variable = $info->{$_}->{hydra_variable};
 
@@ -353,14 +349,12 @@ sub get_query_pattern {
     return undef unless $pattern->{rdf_predicate};
     return undef unless $pattern->{rdf_object};
 
-    $pattern->{hydra_template} = $info->{hydra_template};
+    $pattern->{void_uriLookupEndpoint} = $info->{void_uriLookupEndpoint};
 
     $pattern;
 }
 
-#----------------------------------------------------------------------------------
-
-# Public method
+# Given $subject,$predicate,$object return a generator for RDF::Trine::Model
 sub get_statements {
     my ($self,@triple) = @_;
     my ($subject,$predicate,$object);
@@ -368,67 +362,54 @@ sub get_statements {
     if (@triple == 3) {
         ($subject,$predicate,$object) = @triple;
     }
-    elsif (_is_hash_ref($triple[0])) {
+    elsif (is_hash_ref($triple[0])) {
         $subject   = $triple[0]->{subject};
         $predicate = $triple[0]->{predicate};
         $object    = $triple[0]->{object};
     }
 
-    $subject   = $subject->value if (_is_blessed($subject) && $subject->isa('RDF::Trine::Node') and not $subject->is_variable);
-    $predicate = $predicate->value if (_is_blessed($predicate) && $predicate->isa('RDF::Trine::Node') and not $predicate->is_variable);
-    if (_is_blessed($object) && $object->isa('RDF::Trine::Node') and not $object->is_variable) {
-        $object = ($object->isa('RDF::Trine::Node::Literal')) ? $object->as_string : $object->value;
-    }
+    my $pattern = $self->pattern;
 
-    # Do a federated search over all the URLs provided
-    my $parts  = ref($self->url) ? $self->url : [ $self->url ];
-    my @federated;
+    return undef unless defined $pattern;
 
-    for my $part (@$parts) {
-        my $pattern = $self->get_query_pattern($part);
-        return undef unless defined $pattern;
+    my @param = ();
 
-        my %params;
-        $params{ $pattern->{rdf_subject} }   = $subject if _is_string($subject);
-        $params{ $pattern->{rdf_predicate} } = $predicate if _is_string($predicate);
-        $params{ $pattern->{rdf_object} }    = $object if _is_string($object);
+    push @param , $pattern->{rdf_subject}   . "=" . uri_escape($subject)   if is_string($subject);
+    push @param , $pattern->{rdf_predicate} . "=" . uri_escape($predicate) if is_string($predicate);
+    push @param , $pattern->{rdf_object}    . "=" . uri_escape($object)    if is_string($object);
 
-        my $template  = URI::Template->new($pattern->{hydra_template});
-        push @federated , $template->process(%params)->as_string;
+    my $url = $self->url;
+
+    if (@param) {
+        my $params = join("&",@param);
+        $url = $pattern->{void_uriLookupEndpoint};
+        $url =~ s/{\?\S+}/?$params/;
     }
 
     my $sub = sub {
         state $model;
         state $info;
         state $iterator;
-        state $url = shift(@federated);
 
-        my $triple;
+        unless (defined $model) {
+            return unless defined $url;
 
-        do {
-            unless (defined $model) {
-                # When no more result pages are available switch
-                # to the next federated url...
-                return unless defined($url) || defined($url = pop(@federated));
+            my $fragment = $self->get_model_and_info($url);
 
-                my $fragment = $self->get_model_and_info($url);
+            return unless defined $fragment->{model};
 
-                return unless defined $fragment->{model};
+            $model    = $fragment->{model};
+            $info     = $fragment->{info};
 
-                $model    = $fragment->{model};
-                $info     = $fragment->{info};
+            $url      = $info->{hydra_nextPage};
+            $iterator = $model->get_statements;
+        }
 
-                $url      = $info->{hydra_next};
-                $iterator = $model->get_statements;
-            }
+        my $triple = $iterator->next;
 
-            $triple = $iterator->next;
-
-            unless ($iterator->peek) {
-                $model = undef;
-            }
+        unless ($iterator->peek) {
+            $model = undef;
         }
-        while (!defined $triple && defined($url = pop(@federated)));
 
         wantarray ? ($triple,$info) : $triple;
     };
@@ -466,24 +447,36 @@ sub get_fragment {
 
     $self->log->info("fetching: $url");
 
-    my $model = RDF::Trine::Model->temporary_model;
+    my $req = GET $url, Accept => 'text/turtle';
 
-    # JSON support in RDF::Trine isn't JSON-LD
-    # Set the accept header quality parameter at a minimum for this format
-    my $ua = clone(RDF::Trine->default_useragent);
-    $ua->agent("RDF:::LDF/$RDF::LDF::VERSION " . $ua->_agent);
-    $ua->default_header('Accept','text/turtle;q=1.0,application/turtle;q=1.0,application/x-turtle;q=1.0,application/rdf+xml;q=0.9,text/x-nquads;q=0.9,application/json;q=0.1,application/x-rdf+json;q=0.1');
+    my $response = $self->ua->request($req);
+
+    if ($response->is_success) {
+        $self->parse_string($response->decoded_content);
+    }
+    else {
+        warn Dumper($response);
+        Catmandu::Error->throw("$url failed");
+    }
+}
+
+# Parse turtle into an RDF::Trine::Model
+sub parse_string {
+    my ($self,$string) = @_;
+    $self->log->debug("parsing: $string");
+    my $parser = RDF::Trine::Parser->new('turtle');
+    my $model  = RDF::Trine::Model->temporary_model;
 
     eval {
-        # Need to explicitly set the useragent to keep the accept headers
-        RDF::Trine::Parser->parse_url_into_model($url, $model, useragent => $ua);
+        $parser->parse_into_model($self->url,$string,$model);
     };
 
     if ($@) {
         $self->log->error("failed to parse input");
+        return undef;
     }
 
-    return $model;
+    $model;
 }
 
 # Create a hash with fragment metadata from a RDF::Trine::Model
@@ -491,7 +484,7 @@ sub get_fragment {
 #    $model    - RDF::Trine::Model
 #    $this_uri - result page URL
 #    %opts
-#        clean => 1 - remove the metadata from the model
+#        clean => 1 - remove the metadata from the model 
 sub _model_metadata {
     my ($self,$model,$this_uri,%opts) = @_;
 
@@ -519,8 +512,7 @@ sub _model_metadata {
         'http://www.w3.org/ns/hydra/core#property' ,
         'http://www.w3.org/ns/hydra/core#mapping'  ,
         'http://www.w3.org/ns/hydra/core#template' ,
-        'http://www.w3.org/ns/hydra/core#member'   ,
-        'http://www.w3.org/ns/hydra/core#variableRepresentation' ,
+        'http://www.w3.org/ns/hydra/core#membe'    ,
     ) {
         $self->_build_metadata($model, {
             predicate => RDF::Trine::Node::Resource->new($predicate)
@@ -534,7 +526,7 @@ sub _model_metadata {
         }
     }
 
-    my $source = $info->{dct_source}->[0] if _is_array_ref($info->{dct_source});
+    my $source = $info->{dct_source}->[0] if is_array_ref($info->{dct_source});
 
     if ($source) {
         $self->_build_metadata($model, {
@@ -561,7 +553,7 @@ sub _model_metadata {
 # Helper method for _parse_model
 sub _build_metadata {
     my ($self, $model, $triple, $info) = @_;
-
+    
     my $iterator = $model->get_statements(
         $triple->{subject},
         $triple->{predicate},
@@ -596,22 +588,6 @@ sub _build_metadata {
     $info;
 }
 
-sub _is_array_ref {
-    return ref($_[0]) eq 'ARRAY';
-}
-
-sub _is_hash_ref {
-    return ref($_[0]) eq 'HASH';
-}
-
-sub _is_blessed {
-    return ref($_[0]) =~ /\S/;
-}
-
-sub _is_string {
-    return defined($_[0]) && !ref($_[0]) && ref(\$_[0]) ne 'GLOB' && length($_[0]) > 0;
-}
-
 1;
 
 __END__
@@ -620,57 +596,35 @@ __END__
 
 RDF::LDF - Linked Data Fragments client
 
-=begin markdown
-
-# STATUS
-[![Build Status](https://travis-ci.org/phochste/RDF-LDF.svg)](https://travis-ci.org/phochste/RDF-LDF)
-[![Coverage Status](https://coveralls.io/repos/phochste/RDF-LDF/badge.svg)](https://coveralls.io/r/phochste/RDF-LDF)
-[![Kwalitee Score](http://cpants.cpanauthors.org/dist/RDF-LDF.png)](http://cpants.cpanauthors.org/dist/RDF-LDF)
-
-=end markdown
-
 =head1 SYNOPSIS
 
-    use RDF::Trine::Store::LDF;
-    use RDF::Trine::Store;
+    use RDF::LDF ;
 
-    # To use a HTTP cache:
-    use LWP::UserAgent::CHICaching;
-    my $cache = CHI->new( driver => 'Memory', global => 1 );
-    my $ua = LWP::UserAgent::CHICaching->new(cache => $cache);
-    RDF::Trine->default_useragent($ua);
-
-    my $store = RDF::Trine::Store->new_with_config({
-            storetype => 'LDF',
-            url => $url
-    });
-
-    my $it = $store->get_statements();
-
-    while (my $st = $it->next) {
-        # $st is a RDF::Trine::Statement
-        print "$st\n";
-    }
-
-    # Or the low level modules themselves
-
-    use RDF::LDF;
-
-    my $client = RDF::LDF->new(url => 'http://fragments.dbpedia.org/2014/en');
+    my $client = RDF::LDF ->new(url => 'http://fragments.dbpedia.org/2014/en');
 
     my $iterator = $client->get_statements($subject, $predicate, $object);
 
     while (my $statement = $iterator->()) {
         # $model is a RDF::Trine::Statement
-    }
+    } 
 
+    my $iterator = $client->get_sparql(<<EOF);
+PREFIX dbpedia: <http://dbpedia.org/resource/>
+SELECT * WHERE { dbpedia:Arthur_Schopenhauer ?predicate ?object . }
+EOF
+
+    while (my $binding = $iterator->()) {
+        # $binding is a hashref of all the bindings in the SPARQL
+    }
 
 =head1 DESCRIPTION
 
-RDF::LDF implements a basic L<Linked Data Fragment|http://linkeddatafragments.org/> client.
+The RDF::LDF  module is a basic implementation of a Linked Data Fragment client. For details see:
+<http://linkeddatafragments.org/>.
 
-This a low level module to implement the Linked Data Fragment protocol. You probably want to
-use L<RDF::Trine::Store::LDF>.
+=head1 STATUS
+
+THIS IS ALPHA CODE! The implementation is unreliable and the interface is subject to change.
 
 =head1 CONFIGURATION
 
@@ -680,47 +634,47 @@ use L<RDF::Trine::Store::LDF>.
 
 URL to retrieve RDF from.
 
-Experimental: more than one URL can be provided for federated search over many LDF endpoints.
-
-    my $store = RDF::Trine::Store->new_with_config({
-            storetype => 'LDF',
-            url => [ $url1, $url2, $url3 ]
-    });
-
-    # or
-
-    my $client = RDF::LDF->new(url => [ $url1, $url2, $url3 ]);
-
 =back
 
 =head1 METHODS
 
-=over
+=over 
 
-=item get_statements( $subject, $predicate, $object )
+=item get_statements($subject,$predicate,$object)
 
 Return an iterator for every RDF::Trine::Statement served by the LDF server.
 
-=item get_pattern( $bgp );
+=item get_sparql($sparql)
 
-Returns a stream object of all bindings matching the specified graph pattern.
+Return an iterator for everty binding in the SPARQL query. For now the support is very basic.
+Only select queries are supported without union, sorting or limits with a form like:
+
+
+    PREFIX owl: <http://www.w3.org/2002/07/owl#>
+    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
+    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
+    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
+    PREFIX foaf: <http://xmlns.com/foaf/0.1/>
+    PREFIX dc: <http://purl.org/dc/elements/1.1/>
+    PREFIX : <http://dbpedia.org/resource/>
+    PREFIX dbpedia2: <http://dbpedia.org/property/>
+    PREFIX dbpedia: <http://dbpedia.org/>
+    PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
+    PREFIX dbo: <http://dbpedia.org/ontology/>
+    SELECT *
+    WHERE {
+      ?car <http://purl.org/dc/terms/subject> <http://dbpedia.org/resource/Category:Luxury_vehicles> .
+      ?car foaf:name ?name .
+      ?car dbo:manufacturer ?man .
+      ?man foaf:name ?manufacturer
+    }
 
 =back
 
-=head1 CONTRIBUTORS
+=head1 AUTHOR
 
 Patrick Hochstenbach, C<< patrick.hochstenbach at ugent.be >>
 
-Gregory Todd Williams, C<< greg@evilfunhouse.com >>
-
-Jacob Voss, C<< voss@gbv.de >>
-
-=head1 COPYRIGHT AND LICENSE
-
-This software is copyright (c) 2015 by Patrick Hochstenbach.
-
-This is free software; you can redistribute it and/or modify it under the same terms as the Perl 5 programming language system itself.
-
 =encoding utf8
 
-=cut
+=cut
\ No newline at end of file
